#include "demo.h"
#include "demo_legacy.h"
#include "demfilter.h"
#include "demarathon.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

/***************************************************************************
 *
 * DATA TYPES
 *
 ***************************************************************************/

/***************************************************************************
 *
 * DEFINES
 *
 ***************************************************************************/

#define MAX_FILENAME_LEN 51

/***************************************************************************
 *
 * FORWARD REFERENCES
 *
 ***************************************************************************/

/***************************************************************************
 *
 * EXPORTED FUNCTIONS
 *
 ***************************************************************************/

/***************************************************************************
 *
 * STATIC FUNCTIONS
 *
 ***************************************************************************/

// helper, checks if a file name is already taken
static int name_busy(struct marathon *m, int max, char *filename)
{
  int i;

  for (i = 0; i < max; i++) {
    if (strcmp(m[i].filename, filename) == 0) {
      return 1;
    }
  }
  return 0;
}

int make_marathon(demo *mydemo, struct marathon *marathon,
                  int max, int *count)
{
  char filename[MAX_FILENAME_LEN];
  block *b;
  message *m;
  struct marathon *d;
  int mcount;
  int democount = 0;
  char *map_p;
  char *mapname;
  int i;

  *count = 0;

  // parse into separate demos
  for (b = mydemo->blocks; b != NULL; b = b->next) {
    for (m = b->messages, mcount = 0;
         m != NULL && mcount < 3;
         m=m->next, mcount++)
      {
        if (m->type == SERVERINFO) {
          // new demo found
          d = &marathon[democount];
          d->demo.track = -1;
          d->demo.blocks = b;

          // extract filename
          for (map_p = m->data + 6; *map_p != 0; map_p++) ;
          map_p += 1;
          mapname = strdup(map_p + 5);
          map_p = mapname;
          do {
            if (*map_p == '.') {
              *map_p = '\0';
              break;
            }
            map_p++;
          } while (*map_p != 0);
          d->mapname = mapname;

          // unlink the demo
          if (d->demo.blocks && d->demo.blocks->prev) {
            d->demo.blocks->prev->next = NULL;
            d->demo.blocks->prev = NULL;
          }

          // account for the new demo
          democount++;
        }
      }
  }

  // make unique file names
  for (i = 0; i < democount; i++) {
    int iterator;
    strncpy(filename, marathon[i].mapname, MAX_FILENAME_LEN);
    if (name_busy(marathon, i - 1, filename)) {
      iterator = 2;
      do {
        sprintf(filename, "%s%d", marathon[i].mapname, iterator);
        iterator++;
      } while (name_busy(marathon, i - 1, filename) && iterator < max);
      if (iterator == max) {
        return MARATHON_NONUNIQUE_DEMO;
      }
    }
    marathon[i].filename = strdup(filename);
  }

  *count = democount;

  return MARATHON_OK;
}

int filter_demos(struct marathon *m, int count)
{
  int i;

  for (i = 0; i < count; i++) {
    char stufftext[100];
    add_nop_block(&m[i].demo);
    remove_grenade_counter(&m[i].demo);
    if (i + 1 < count) {
      sprintf(stufftext, "playdemo %s\n", m[i + 1].filename);
      replace_reconnect(&m[i].demo, stufftext);
    }
    m[i].time = demo_time(&m[i].demo);
  }

  return FILTER_OK;
}

int make_pak(char *pakname, struct marathon *marathon, int count)
{
  FILE * pak;
  struct pak_header header;
  struct pak_dirent dirent;
  int i;
  unsigned int *sizes;
  unsigned int *fposs;
  dret_t dret;
  int ret;

  //open file
#ifdef NO_PAK_OVERWRITE
  pak = fopen(pakname, "r");
  if (pak != NULL) {
    fclose(pak);
    return PAK_FILE_EXISTS;
  }
#endif
  pak = fopen (pakname, "wb");
  if(pak == NULL) {
    return PAK_CANNOT_OPEN;
  }

  sizes = (unsigned int *)malloc(count * sizeof(unsigned int));
  fposs = (unsigned int *)malloc(count * sizeof(unsigned int));

  if (sizes == NULL || fposs == NULL) {
    return PAK_NO_MEMORY;
  }

  // write dummy header
  memset(&header, 0, sizeof(header));
  fwrite(&header, sizeof(header), 1, pak);

  // write all demo data
  for (i = 0; i < count; i++) {
    fposs[i] = ftell(pak);
    dret = write_demo_fp(pak, &marathon[i].demo);
    if (ret != DEMO_OK) {
      switch (dret) {
      case DEMO_OK:
        break;

      case DEMO_CORRUPT_DEMO:
        ret = PAK_CORRUPT_DEMO;
        goto make_pak_done;

      default:
        ret = PAK_CANNOT_WRITE;
        goto make_pak_done;
      }
    }
    sizes[i] = ftell(pak) - fposs[i];
  }

  // get directory offset
  header.doffs = htodl(ftell(pak));

  // write directory
  for (i = 0; i < count; i++) {
    memset(&dirent, 0, sizeof(dirent));
    sprintf(dirent.filename, "%s.dem", marathon[i].filename);
    dirent.fpos = htodl(fposs[i]);
    dirent.flen = htodl(sizes[i]);
    fwrite(&dirent, sizeof(dirent), 1, pak);
  }

  // update header
  fseek(pak, 0, SEEK_SET);
  header.dlen = htodl(count * sizeof(struct pak_dirent));
  header.sig[0] = 'P';
  header.sig[1] = 'A';
  header.sig[2] = 'C';
  header.sig[3] = 'K';
  fwrite(&header, sizeof(header), 1, pak);

 make_pak_done:
  fclose(pak);
  free(sizes);
  free(fposs);
  return PAK_OK;
}

char *pak_error(int errcode)
{
  switch (errcode) {
  case PAK_OK:
    return "no error";

  case PAK_FILE_EXISTS:
    return "pak file exists";

  case PAK_CANNOT_OPEN:
    return "cannot open pak file";

  case PAK_NO_MEMORY:
    return "ran out of memory";

  case PAK_CANNOT_WRITE:
    return "file write error";

  case PAK_CORRUPT_DEMO:
    return "corrupt demo data";

  default:
    return "unknown pak error";
  }
}

char *marathon_error(int errcode)
{
  switch (errcode) {
  case MARATHON_OK:
    return "no error";

  case MARATHON_NONUNIQUE_DEMO:
    return "couldn't make unique file name for a demo";

  default:
    return "unknown marathon error";
  }
}

char *filter_error(int errcode)
{
  switch(errcode) {
  case FILTER_OK:
    return "no error";

  default:
    return "unknown filter error";
  }
}

extern char *demarathon_error(int errcode)
{
  static char errbuf[64];
  if (errcode >= PAK_EBASE) {
    return pak_error(errcode);
  }
  else if (errcode >= FILTER_EBASE) {
    return filter_error(errcode);
  }
  else if (errcode >= MARATHON_EBASE) {
    return marathon_error(errcode);
  }
  else {
    sprintf(errbuf, "unknown error %d\n", errcode);
    return errbuf;
  }
}
