#include "demo.h"
#include "stdio.h"
#include "string.h"
#include "stdlib.h"
#include "demfilter.h"

static int is_empty_block(block *b);
static int remove_block(demo *d, block *b);
static int remove_message(block *b, message *mess);

char *demo_time(demo *d)
{
  block *b;
  message *m;
  char *time = NULL;
  char timearray[128] = {0};

  for (b = d->blocks; b; b = b->next) {
    for (m = b->messages; m; m = m->next) {
      if (m->type == PRINT) {
        if (strstr(m->data, "The recorded time was ") != NULL) {
          for (m = m->next;
               m && m->type == PRINT && strcmp(m->data, "\n") != 0;
               m = m->next)
          {
            strcat(timearray, m->data);
          }
          time = strdup(timearray);
          break;
        }
      }
    }
  }

  return time;
}

// removes grenade counter
int remove_grenade_counter(demo *d)
{
  block *b, *bnext;
  message *m, *mnext;
  int removed = 0;

  for (b = d->blocks; b != NULL; b = bnext) {
    bnext = b->next;
    m = b->messages;
    while (m) {
      mnext = m->next;
      if (m->type == CENTERPRINT &&
          (!strncmp(m->data, "Grenade ", 8) ||
           !strncmp(m->data, "Ogre-Grenade", 12)))
      {
        remove_message(b, m);
        removed++;
      }
      if (is_empty_block(b)) {
        remove_block(d, b);
      }
      m = mnext;
    }
  }

  return removed;
}

// adds a nop block to the start of the demo
int add_nop_block(demo *d)
{
  block *b;

  if (d->blocks &&
      d->blocks->messages &&
      d->blocks->messages->type == NOP)
  {
    return 0;
  }

  b = (block *)calloc(1, sizeof(block));
  b->length = 1;
  b->messages = (message *)calloc(1, sizeof(message));
  b->messages->type = NOP;
  if (d->blocks) {
    b->next = d->blocks;
    d->blocks->prev = b;
  }
  d->blocks = b;

  return 1;
}

// removes reconnect stufftext if present
int replace_reconnect(demo *d, char *stufftext)
{
  block *b;
  message *m;
  int deficit;

  for (b = d->blocks; b != NULL; b = b->next) {
    for (m = b->messages; m != NULL; m = m->next) {
      // filter here
      if (m->type == STUFFTEXT) {
        if (strcmp(m->data, "reconnect\n") == 0) {
          int newlen;

          // found our entry point
          deficit = m->size;
          m->size -= strlen(m->data);
          free(m->data);

          newlen = strlen(stufftext) + 2;
          m->data = calloc(newlen, 1);
          strcpy(m->data, stufftext);
          m->data[newlen - 1] = DISCONNECT;
          m->size += newlen - 1;

          deficit -= m->size;
          b->length -= deficit;

          return 1;
        }
      }
    }
  }
  return 1;
}

static int is_empty_block(block *b)
{
  return !b->length;
}

static int remove_block(demo *d, block *b)
{
  message *m;
  message *mnext;

  if (d->blocks == b) {
    d->blocks = b->next;
    if (b->next) {
      b->next->prev = NULL;
    }
  } else if (b->prev) {
    b->prev->next = b->next;
    if (b->next) {
      b->next->prev = b->prev;
    }
  } else {
    return -1;
  }

  for (m = b->messages; m != NULL; m = mnext) {
    mnext = m->next;
    free(m->data);
    free(m);
  }
  free(b);

  return 0;
}

static int remove_message(block *b, message *mess)
{
  message *m;
  m = b->messages;

  for (m = b->messages; m != NULL && m != mess; m = m->next);

  if (m == NULL) {
    return -1;
  }

  // unlink
  if (m->prev != NULL) {
    m->prev->next = m->next;
    if (m->next != NULL) {
      m->next->prev = m->prev;
    }
  }
  else {
    b->messages = m->next;
    if (m->next) {
      m->next->prev = NULL;
    }
  }

  b->length -= (m->size + 1);
  free(m->data);
  free(mess);

  return 0;
}
